import click
import sys
import os
from importlib import import_module
from functools import update_wrapper

from . import __version__
from . import Client


def main_wrapper(cmd, args, **kwargs):
    if args is None:
        args = sys.argv[1:]

    def decorator():
        try:
            cmd.main(args=args, **kwargs)
        except Exception as e:  # pragma: no cover
            if os.environ.get("HELIUM_COMMANDER_DEBUG"):
                raise
            click.secho(str(e), fg='red')
            sys.exit(1)
    return decorator


CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)


def main_commands(version=None, package=None,  commands=None, name='helium'):
    class Loader(click.MultiCommand):
        def list_commands(self, ctx):
            commands.sort()
            return commands

        def get_command(self, ctx, name):
            try:
                command = import_module(package + "." + name)
                return command.cli
            except ImportError as e:  # pragma: no cover
                click.secho(str(e), fg='red')
                return

    def decorator(f):
        @click.option('--uuid', is_flag=True,
                      help="Whether to display long identifiers")
        @click.option('--format',
                      type=click.Choice(['csv', 'json', 'tabular']),
                      default=None,
                      help="The output format (default 'tabular')")
        @click.option('--output',
                      type=click.File('w'),
                      help="the destination file for results")
        @click.version_option(version=version)
        @click.command(name, cls=Loader, context_settings=CONTEXT_SETTINGS)
        @click.pass_context
        def new_func(ctx, *args, **kwargs):
            ctx.invoke(f, ctx, *args, **kwargs)
        return update_wrapper(new_func, f)
    return decorator


_commands = [
    "user",
    "sensor",
    "element",
    "label",
    "organization",
]


@click.option('--api-key',
              envvar='HELIUM_API_KEY',
              help='your Helium API key. Can also be specified using the HELIUM_API_KEY environment variable')
@click.option('--host',
              envvar='HELIUM_API_URL',
              default='https://api.helium.com/v1',
              help='The Helium base API URL. Can also be specified using the HELIUM_API_URL environment variable.' )
@main_commands(version=__version__, package='helium_commander.commands', commands=_commands)
def root(ctx, api_key, host, **kwargs):
    ctx.obj = Client(api_token=api_key, base_url=host, **kwargs)


main = main_wrapper(root, None)


if __name__ == '__main__':      # pragma: no cover
    main()



from __future__ import unicode_literals
import sys
import json
import abc
import terminaltables
from textwrap import wrap
from operator import itemgetter
from future.utils import iteritems
if sys.version_info[0] == 2:    # pragma: no cover
    import unicodecsv as csv
else:                           # pragma: no cover
    import csv


def for_format(format, file):
    json_opts = None
    if format == 'json':
        json_opts = json_opts or {
            "indent": 4
        }
        return JSONWriter(file, json_opts=json_opts)
    elif format == 'csv':
        json_opts = json_opts or {
            "indent": None
        }
        return CSVWriter(file, json_opts=json_opts)
    elif format == 'tabular':
        json_opts = json_opts or {
            "indent": 0
        }
        return TabularWriter(file, json_opts=json_opts)
    else:
        raise AttributeError("Unrecognized format {}".format(format))


class BaseWriter:
    __metaclass__ = abc.ABCMeta

    def __init__(self, file, json_opts={}):
        self.file = file
        self.json_opts = json_opts

    def start(self, mapping, **kwargs):
        self.sort = kwargs.pop('sort', None)
        self.reverse = kwargs.pop('reverse', False)

    @abc.abstractmethod
    def write_resources(self, resources):
        """Abstract method for subclasses to implement"""

    def finish(self, mapping):
        return

    def lookup(self, o, path):
        result = path(o)
        if result is None:
            return ''
        if isinstance(result, dict) and self.json_opts:
            result = json.dumps(result, **self.json_opts)
        else:
            result = result
        return result

    def order_entries(self, entries, keyfunc):
        if self.sort and keyfunc:
            entries = sorted(entries,
                             key=keyfunc,
                             reverse=self.reverse)
        elif self.reverse:
            entries = reversed(list(entries))
        return entries


class CSVWriter(BaseWriter):

    def start(self, mapping, **kwargs):
        super(CSVWriter, self).start(mapping, **kwargs)
        self.writer = csv.DictWriter(self.file, mapping.keys())
        self.writer.writeheader()

    def write_resources(self, entries, mapping):
        def map_entry(o):
            return {key: self.lookup(o, path)
                    for key, path in iteritems(mapping)}

        entries = map(map_entry, entries)

        keyfunc = itemgetter(self.sort) if self.sort else None
        entries = self.order_entries(entries, keyfunc)
        for entry in entries:
            self.writer.writerow(entry)


class JSONWriter(BaseWriter):

    def start(self, mapping, **kwargs):
        super(JSONWriter, self).start(mapping, **kwargs)
        self.file.write('[\n')
        self.is_first_entries = True

    def write_resources(self, entries, mapping):
        def map_entry(o):
            return {key: self.lookup(o, path)
                    for key, path in iteritems(mapping)}

        entries = map(map_entry, entries)
        keyfunc = itemgetter(self.sort) if self.sort else None
        entries = self.order_entries(entries, keyfunc)

        for entry in entries:
            if self.is_first_entries:
                self.is_first_entries = False
            else:
                self.file.write(',')
            json.dump(entry, self.file, **self.json_opts)

    def finish(self, mapping):
        self.file.write('\n]\n')


class TabularWriter(BaseWriter):

    def start(self, mapping, **kwargs):
        super(TabularWriter, self).start(mapping, **kwargs)
        self.data = [[key.upper() for key in mapping.keys()]]
        self.max_width = kwargs.get('max_width', None)

    def write_resources(self, resources, mapping):
        def map_entry(o):
            return [self.lookup(o, path) for _, path in mapping.items()]

        entries = [map_entry(o) for o in resources]
        if self.sort:
            sort_index = list(mapping.keys()).index(self.sort)
            keyfunc = itemgetter(sort_index)
        else:
            keyfunc = None
        entries = self.order_entries(entries, keyfunc)
        self.data.extend(entries)

    def finish(self, mapping):
        table = terminaltables.AsciiTable(self.data)
        last_column = len(mapping) - 1
        max_width = self.max_width or table.column_max_width(last_column)
        for entry in table.table_data:
            last_field = entry[last_column]
            # check for basic wrap functionality
            if hasattr(last_field, 'expandtabs'):
                entry[last_column] = '\n'.join(wrap(last_field, max_width))
        self.file.write(table.table)
        self.file.write('\n')



"""Extensions to the base Resource class."""

from __future__ import unicode_literals

from helium import Resource, ResourceMeta
import helium_commander.writer as _writer
import click
from builtins import filter as _filter
from collections import OrderedDict
from contextlib import contextmanager


def filter_mac(id_rep):
    id_rep_lower = id_rep.lower()

    def func(resource):
        try:
            resource_mac = resource.meta.mac
            return resource_mac[-len(id_rep):].lower() == id_rep_lower
        except AttributeError:
            pass
        return False
    return func


def filter_attribute(attribute):
    def _filter(id_rep):
        def func(resource):
            return id_rep == getattr(resource, attribute, None)
        return func
    return _filter


def filter_string_attribute(attribute):
    def _filter_attribute(id_rep):
        id_rep_lower = None
        id_rep_len = 0
        if id_rep is not None:
            id_rep_lower = id_rep.lower()
            id_rep_len = len(id_rep_lower)

        def func(resource):
            value = getattr(resource, attribute, None)
            if value is None:
                return id_rep_lower is None
            return value[:id_rep_len].lower() == id_rep_lower
        return func
    return _filter_attribute


def filter_oneof(lookup_filters):
    def id_func(id_rep):
        resource_filters = [f(id_rep) for f in lookup_filters]

        def func(resource):
            for f in resource_filters:
                if f(resource):
                    return True
            return False
        return func
    return id_func


def filter_id_rep(mac=False):
    if mac:
        filters = [
            filter_mac
        ]
    else:
        filters = [
            filter_attribute('id'),
            filter_attribute('short_id'),
            filter_string_attribute('name')
        ]
    return filter_oneof(filters)


def resource_filter(cls, client, value,
                    lookup_filter=None,
                    resources=None,
                    include=None):
    if resources is None:
        resources = cls.all(client, include=include)
    lookup_filter = lookup_filter or filter_attribute('id')
    resource_filter = lookup_filter(value)
    return list(_filter(resource_filter, resources))


def resource_lookup(cls, client, id_rep,
                    mac=False,
                    resources=None,
                    lookup_filter=None,
                    include=None):
    lookup_filter = lookup_filter or filter_id_rep(mac=mac)
    resources = resource_filter(cls, client, id_rep,
                                lookup_filter=lookup_filter,
                                resources=resources,
                                include=include)
    resources_len = len(list(resources)) if resources is not None else 0
    if resources_len == 1:
        return resources[0]
    elif resources_len == 0:
        raise KeyError('Id: {} does not exist'.format(id_rep))
    else:
        ambig_ids = [res.id for res in resources]
        ambig_list = '({})'.format(', '.join(ambig_ids))
        raise KeyError('Ambiguous id: {} {}'.format(id_rep, ambig_list))


Resource.filter = classmethod(resource_filter)
Resource.lookup = classmethod(resource_lookup)
Resource.short_id = property(lambda self: self.id.split('-')[0])


def display_map(cls, client, uuid=False):
    def _id(self):
        return self.id if uuid or client.uuid else self.short_id

    return OrderedDict([
        ('id', _id),
    ])


@contextmanager
def display_writer(cls, client, mapping, **kwargs):
    output = kwargs.get('output', client.output)
    output = output or click.utils.get_text_stream('stdout')
    output_format = kwargs.get('format', client.format)
    writer = _writer.for_format(output_format, output)
    writer.start(mapping, **kwargs)
    try:
        yield writer
    finally:
        writer.finish(mapping)


def display_resources(cls, client, resources, **kwargs):
    mapping = cls.display_map(client)
    with cls.display_writer(client, mapping, **kwargs) as writer:
        writer.write_resources(resources, mapping)

Resource.display_writer = classmethod(display_writer)
Resource.display_map = classmethod(display_map)
Resource.display = classmethod(display_resources)

ResourceMeta                    # Avoid unused warning



from __future__ import unicode_literals
from helium import DataPoint, Timeseries
from operator import attrgetter


def display_map(cls, client):
    dict = super(DataPoint, cls).display_map(client)
    dict.update([
        ('timestamp', attrgetter('timestamp')),
        ('port', attrgetter('port')),
        ('value', attrgetter('value')),
    ])
    return dict

DataPoint.display_map = classmethod(display_map)
Timeseries                      # Empty reference to make clear we re-export



from __future__ import unicode_literals
from helium import User
from helium import response_json


def display_map(cls, client):
    def _name(self):
        return getattr(self, 'name', None)

    def _email(self):
        """Display the email or the id.

        The public API uses the email as the id. So display that if
        email can not be found for some weird reason.
        """
        return getattr(self, 'email', getattr(self, 'id'))

    dict = super(User, cls).display_map(client, uuid=False)
    dict.update([
        ('email', _email),
        ('name', _name),
    ])
    return dict


def authenticate(cls, client, email, password):   # pragma: no cover
    url = client._build_url('user', 'auth')
    json = {
        'email': email,
        'password': password
    }
    json = response_json(client.post(url, json=json), 200, extract=None)
    return cls._mk_one(client, json)

User.display_map = classmethod(display_map)
User.authenticate = classmethod(authenticate)



from __future__ import unicode_literals
from helium import Organization
from operator import attrgetter


def display_map(cls, client):
    def _count_user(self):
        return len(self.users(use_included=True))

    def _count_element(self):
        return len(self.elements(use_included=True))

    def _count_sensor(self):
        return len(self.sensors(use_included=True))

    dict = super(Organization, cls).display_map(client)
    dict.update([
        ('users', _count_user),
        ('element', _count_element),
        ('sensor', _count_sensor),
        ('name', attrgetter('name')),
    ])
    return dict

Organization.display_map = classmethod(display_map)



from __future__ import unicode_literals

import helium


class Client(helium.Client):

    def __init__(self, **kwargs):
        token = kwargs.pop("api_token", None)
        url = kwargs.pop('base_url', None)
        super(Client, self).__init__(api_token=token, base_url=url)

        self.uuid = kwargs.get('uuid', None) or False
        self.format = kwargs.get("format", None) or 'tabular'
        self.output = kwargs.get('output', None)



import click
from helium_commander import Client, Organization
from helium_commander import User, Element, Sensor
from helium_commander.commands import timeseries


pass_client = click.make_pass_decorator(Client)


@click.group()
def cli():
    """Operations on the authorized organization
    """
    pass


@cli.command()
@pass_client
def list(client):
    """Display basic information of the organization.

    Displays basic attributes of the authorized organization.
    """
    org = Organization.singleton(client, include=[User, Element, Sensor])
    Organization.display(client, [org])


@cli.command()
@click.option('--name',
              help="the new name for the organization")
@pass_client
def update(client, name):
    """Updates the attributes of the organization.

    Updates the attributes of the currently authorized organization.
    """
    org = Organization.singleton(client)
    org = org.update(name=name)
    org = Organization.singleton(client, include=[User, Element, Sensor])
    Organization.display(client, [org])


cli.add_command(timeseries.cli(Organization, singleton=True))



from __future__ import unicode_literals
from helium import Element


def display_map(cls, client):
    def _mac(self):
        return self.meta.mac

    def _count_sensor(self):
        try:
            return len(self.sensors(use_included=True))
        except AttributeError:
            pass
        return None

    def _seen(self):
        try:
            return self.meta.last_seen
        except AttributeError:
            pass
        return None

    def _name(self):
        return getattr(self, 'name', None)

    dict = super(Element, cls).display_map(client)
    dict.update([
        ('mac', _mac),
        ('sensors', _count_sensor),
        ('seen', _seen),
        ('name', _name)
    ])
    return dict

Element.display_map = classmethod(display_map)



import click
from helium_commander import Client, Label, Sensor
from helium_commander import device_sort_option, ResourceParamType

pass_client = click.make_pass_decorator(Client)


@click.group()
def cli():
    """Operations on labels of sensors.
    """
    pass


@cli.command()
@click.argument('label', required=False)
@pass_client
def list(client, label, **kwargs):
    """List labels.

    Lists information for a given LABEL or all labels in the
    organization.

    """
    if label:
        labels = [Label.lookup(client, label, include=[Sensor])]
    else:
        labels = Label.all(client, include=[Sensor])
    Label.display(client, labels, **kwargs)


@cli.command()
@click.argument('label')
@device_sort_option
@pass_client
def sensor(client, label, **kwargs):
    """Lists sensors for a label.

    Lists sensors for a given LABEL.
    """
    label = Label.lookup(client, label, include=[Sensor])
    sensors = label.sensors(use_included=True)
    Sensor.display(client, sensors, **kwargs)


@cli.command()
@click.argument('name')
@click.option('--add',
              type=ResourceParamType(metavar='SENSOR'),
              help="Add sensors to a label")
@click.pass_context
def create(ctx, name, **kwargs):
    """Create a label.

    Creates a label with a given NAME and an (optional) list of sensors
    associated with that label.
    """
    client = ctx.find_object(Client)
    label = Label.create(client, name=name)

    ctx.invoke(update, client, label=label.id, **kwargs)


@cli.command()
@click.argument('label', nargs=-1)
@pass_client
def delete(client, label):
    """Delete one or more labels.

    Deletes the LABELs with the given ids
    """
    all_labels = Label.all(client)
    label = [Label.lookup(client, id, resources=all_labels) for id in label]
    for entry in label:
        result = entry.delete()
        click.echo("Deleted: " + entry.id if result else result)


@cli.command()
@click.argument('label')
@click.option('--add',
              type=ResourceParamType(metavar='SENSOR'),
              help="Add sensors to a label")
@click.option('--remove',
              type=ResourceParamType(metavar='SENSOR'),
              help="Remove sensors from a label")
@click.option('--name',
              help="the new name for the label")
@pass_client
def update(client, label, name, **kwargs):
    label = Label.lookup(client, label)
    if name:
        label.update(name=name)

    all_sensors = Sensor.all(client)
    add_sensors = kwargs.pop('add', None) or []
    remove_sensors = kwargs.pop('remove', None) or []
    sensors = [Sensor.lookup(client, s, resources=all_sensors)
               for s in add_sensors]
    if sensors:
        label.add_sensors(sensors)

    sensors = [Sensor.lookup(client, s, resources=all_sensors)
               for s in remove_sensors]
    if sensors:
        label.remove_sensors(sensors)

    label = Label.find(client, label.id, include=[Sensor])
    Label.display(client, [label])



from setuptools import setup

author = 'Helium'
author_email = 'hello@helium.com'
install_requires = [
    'helium-python>=0.1.8',
    'future>=0.15',
    'dpath>=1.4',
    'futures>=3.0',
    'terminaltables>=2.1.0',
    'click>=6.6',
    'unicodecsv>=0.14.1',
]
setup_requires = [
    'vcversioner',
]

with open('README.md', 'r') as infile:
    long_description = infile.read()

setup(
    name='helium-commander',
    url='http://github.com/helium/helium-commander/',
    author=author,
    author_email=author_email,
    long_description=long_description,
    license='BSD',
    description='A CLI and service wrapper for the Helium API',
    packages=['helium_commander', 'helium_commander.commands'],
    platforms='all',
    install_requires=install_requires,
    setup_requires=setup_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Topic :: Utilities",
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    vcversioner={
        "version_module_paths": ["helium_commander/_version.py"]
    },
    entry_points='''
        [console_scripts]
        helium=helium_commander.cli:main
    '''
)



import click
from functools import reduce
from json import loads as load_json


def sort_option(choices):
    options = [
        click.option('--reverse', is_flag=True,
                     help='Sort in reverse order'),
        click.option('--sort', type=click.Choice(choices),
                     help='How to sort the result')
    ]

    def wrapper(func):
        for option in reversed(options):
            func = option(func)
        return func
    return wrapper


def device_sort_option(f):
    return sort_option(['seen', 'name'])(f)


def device_mac_option(f):
    return click.option('--mac', is_flag=True,
                        help="Whether the given id is a mac address")(f)


class ResourceParamType(click.ParamType):
    name = 'resource'

    def __init__(self, metavar='TEXT'):
        self.metavar = metavar

    def get_metavar(self, param):
        return '{0}[,{0},...]* | @filename'.format(self.metavar)

    def convert(self, value, param, ctx):
        def collect_resources(acc, resource_rep):
            if resource_rep.startswith('@'):
                for line in click.open_file(resource_rep[1:]):
                    acc.append(line.strip())
            else:
                acc.append(resource_rep)
            return acc
        value = value.split(',')
        return reduce(collect_resources, value, [])

    def __repr__(self):
        'Resource(metavar={})'.fomat(self.metavar)


class JSONParamType(click.ParamType):
    name = 'JSON'

    def convert(self, value, param, ctx):
        try:
            return load_json(value)
        except ValueError:
            self.fail('{} is not a valid json value'.format(value), param, ctx)
